
#pragma once

#include "element.h"
#include "element_list.h"

element replace(element e, std::vector<element> const & substitution)
{
   if (e.is_combinator())
      return e;
   else if (e.is_ref())
   {
      element_list * src = e.as_ref();
      element_list * dst = new element_list();
      element ne = element::ref(dst);

      for (element_list::const_iterator i = src->begin(); i != src->end(); ++i)
         dst->push(replace(*i, substitution));

      return ne;
   }
   else if (e.is_argument())
   {
      unsigned a = e.as_argument();
      assert(a < substitution.size());
      return substitution[a];
   }
   else
      throw std::logic_error("bad element type");
}

struct combinator
{
   struct reduction_kind
   {
      reduction_kind(unsigned number_of_argument_to_reduce, element reduction_pattern)
         : number_of_argument_to_reduce(number_of_argument_to_reduce)
         , reduction_pattern(reduction_pattern)
      {}

      unsigned number_of_argument_to_reduce;
      element reduction_pattern;
   };

   combinator()
   {}

   combinator(unsigned number_of_argument_to_reduce, element reduction_pattern)
   {
      reduction_kinds_.push_back(reduction_kind(number_of_argument_to_reduce, reduction_pattern));
   }

   combinator(combinator const & c)
      : reduction_kinds_(c.reduction_kinds_)
   {}

   void add_reduction_kind(reduction_kind const & rk)
   {
      assert(reduction_kinds_.empty() || reduction_kinds_.back().number_of_argument_to_reduce < rk.number_of_argument_to_reduce);
      reduction_kinds_.push_back(rk);
   }

   combinator & operator=(combinator const & c)
   {
      combinator t(c);
      swap(t);

      return *this;
   }

   void swap(combinator & c)
   {
      std::swap(reduction_kinds_, c.reduction_kinds_);
   }

   bool check_for_apply(element_list & e) const
   {
      unsigned c = e.size();

      reduction_kinds_t::const_iterator i = reduction_kinds_.begin(); // TODO: replace with binary search
      if (i == reduction_kinds_.end())
         return false;

      if (i->number_of_argument_to_reduce > c)
         return false;

      return true;
   }

   void apply(element_list & e) const
   {
      unsigned c = e.size();

      reduction_kinds_t::const_iterator i = reduction_kinds_.begin(); // TODO: replace with binary search
      assert(i != reduction_kinds_.end());
      assert(i->number_of_argument_to_reduce <= c);

      for (;;)
      {
         reduction_kinds_t::const_iterator next_i = boost::next(i);
         if (next_i == reduction_kinds_.end())
            break;
         if (next_i->number_of_argument_to_reduce > c)
            break;
         i = next_i;
      }

      std::vector<element> substitution;
      for (unsigned u = 0; u < i->number_of_argument_to_reduce; ++u)
      {
         substitution.push_back(e.top());
         e.pop();
      }

      e.push(replace(i->reduction_pattern, substitution));
   }

private:
   typedef std::vector<reduction_kind> reduction_kinds_t;

public:
   typedef reduction_kinds_t::const_iterator const_iterator;

   const_iterator begin() const
   {
      return reduction_kinds_.begin();
   }

   const_iterator end() const
   {
      return reduction_kinds_.end();
   }

private:
   reduction_kinds_t reduction_kinds_;
};
