
#pragma once

struct element_list // TODO: use it by smart ptr like element, do not use directly
{
private:
   typedef std::vector<element> elements_t;

public:
   typedef elements_t::iterator               iterator;
   typedef elements_t::const_iterator         const_iterator;
   typedef elements_t::reverse_iterator       reverse_iterator;
   typedef elements_t::const_reverse_iterator const_reverse_iterator;

   element_list()
      : ref_count_(0)
   {}

   element_list(element_list const & r)
      : ref_count_(0)
      , elements_(r.elements_)
   {}

   void add_ref()
   {
      ++ref_count_;
   }

   void release()
   {
      --ref_count_;
      if (ref_count_ == 0)
         delete this;
   }

   bool empty() const
   {
      return elements_.empty();
   }

   size_t size() const
   {
      return elements_.size();
   }

   element top() const
   {
      return elements_.back();
   }

   void pop()
   {
      return elements_.pop_back();
   }

   void push(element e)
   {
      elements_.push_back(e);
   }

   iterator begin()
   {
      return elements_.begin();
   }

   const_iterator begin() const
   {
      return elements_.begin();
   }

   iterator end()
   {
      return elements_.end();
   }

   const_iterator end() const
   {
      return elements_.end();
   }

   reverse_iterator rbegin()
   {
      return elements_.rbegin();
   }

   const_reverse_iterator rbegin() const
   {
      return elements_.rbegin();
   }

   reverse_iterator rend()
   {
      return elements_.rend();
   }

   const_reverse_iterator rend() const
   {
      return elements_.rend();
   }

private:
   element_list & operator=(element_list const &);

private:
   unsigned ref_count_;
   elements_t elements_;
};
