
#pragma once

namespace generate_combinator_impl
{
   void check_for_validity(element_list const & el) // TODO: use that as assert in combinator ctor
   {
      for (element_list::const_iterator i = el.begin(); i != el.end(); ++i)
      {
         if (i->is_argument())
            throw std::logic_error("generate_combinator takes expression with constants");
         else if (i->is_ref())
            check_for_validity(*i->as_ref());
         else if (i->is_combinator())
            ;
         else
            throw std::logic_error("bad element type");
      }
   }

   void provide_args(unsigned from, unsigned n, element_list *& clone, element & clone_hold)
   {
      if (n == 0)
         throw std::invalid_argument("can't provide 0 args");

      element_list * new_el = new element_list();

      for (unsigned u = n - 1; (int)u >= 0; --u)
         new_el->push(element::argument(u + from));

      for (element_list::const_iterator i = clone->begin(); i != clone->end(); ++i)
         new_el->push(*i);

      clone = new_el;
      clone_hold = element::ref(clone);
   }
}

combinator generate_combinator(combinator_list const & clist, element_list const & el)
{
   generate_combinator_impl::check_for_validity(el);

   element_list * clone = new element_list(el);
   element clone_hold = element::ref(clone);

   clist.full_reduce(*clone);
//   combinator c(0, element::ref(new element_list(*clone)));
   combinator c;

   unsigned args_provided = 0;
   for (;;)
   {
      if (clone->top().is_argument())
         return c;

      if (clone->top().is_ref())
         throw std::logic_error("oops: full_reduce returns element_list with first element ref");

      if (!clone->top().is_combinator())
         throw std::logic_error("bad element type");

      combinator const & top_c = clist[clone->top().as_combinator()];

      unsigned clone_size = clone->size();
      unsigned nr = top_c.begin()->number_of_argument_to_reduce;

      if ((nr + 1) <= clone_size)
         throw std::logic_error("incomplete reduction");

      unsigned add_arg_count = nr + 1 - clone_size;
      generate_combinator_impl::provide_args(args_provided, add_arg_count, clone, clone_hold);
      args_provided += add_arg_count;

      clist.full_reduce(*clone);
      c.add_reduction_kind(combinator::reduction_kind(args_provided, element::ref(new element_list(*clone))));
   }
}
