
#pragma once

#include "element.h"
#include "element_list.h"

struct abstraction_elimination_context
{
   lambda_item_ptr i;
   lambda_item_ptr k;
   lambda_item_ptr s;
};

struct convert_to_element_context
{
   unsigned i;
   unsigned k;
   unsigned s;
};

lambda_item_ptr abstraction_elimination(abstraction_elimination_context const & c, lambda_item_ptr i)
{
   if (variable * v = dynamic_cast<variable *>(i.get()))
   {
      return i;
   }
   else if (application * app = dynamic_cast<application *>(i.get()))
   {
      return new application(abstraction_elimination(c, app->function_),
                             abstraction_elimination(c, app->arg_));
   }
   else if (abstraction * abs = dynamic_cast<abstraction *>(i.get()))
   {
      if (!abs->body_->is_free_variable(abs->var_->name_))
      {
         return new application(c.k,
                                abstraction_elimination(c, abs->body_));
      }
      else if (variable * v2 = dynamic_cast<variable *>(abs->body_.get()))
      {
         assert(v2->name_ == abs->var_->name_);
         return c.i;
      }
      else if (application * app2 = dynamic_cast<application *>(abs->body_.get()))
      {
         return new application(
            new application(
               c.s,
               abstraction_elimination(
                  c,
                  new abstraction(abs->var_, app2->function_)
               )
            ),
            abstraction_elimination(
               c,
               new abstraction(abs->var_, app2->arg_)
            )
         );
      }
      else if (abstraction * abs2 = dynamic_cast<abstraction *>(abs->body_.get()))
      {
         return abstraction_elimination(
            c,
            new abstraction(
               abs->var_,
               abstraction_elimination(
                  c,
                  new abstraction(
                     abs2->var_,
                     abs2->body_
                  )
               )
            )
         );
      }
      else
         throw std::logic_error("bad lambda_item type");
      
   }
   else
      throw std::logic_error("bad lambda_item type");
}

element convert_to_element(abstraction_elimination_context const & ctx, convert_to_element_context const & ctx2, lambda_item_ptr i)
{
   if (i == ctx.i)
      return element::combinator(ctx2.i);
   else if (i == ctx.k)
      return element::combinator(ctx2.k);
   else if (i == ctx.s)
      return element::combinator(ctx2.s);

   if (variable * v = dynamic_cast<variable *>(i.get()))
      return element::argument(boost::lexical_cast<unsigned>(v->name_));

   element_list * el = new element_list();
   element r(element::ref(el));
   lambda_item_ptr c = i;
   while (application * app = dynamic_cast<application *>(c.get()))
   {
      el->push(convert_to_element(ctx, ctx2, app->arg_));
      c = app->function_;
   }

   {
      variable * v2 = dynamic_cast<variable *>(c.get());
      el->push(convert_to_element(ctx, ctx2, c));
   }

   return r;
}
