
#pragma once


bool is_alpha_num(wchar_t c)
{
   return  (c >= L'A' && c <= L'Z' ||
            c >= L'a' && c <= L'z' ||
            c >= L'0' && c <= L'9');
}

bool is_whitespace(wchar_t c)
{
   return c <= L' ';
}

struct tokenizer
{
   typedef boost::function<boost::optional<wchar_t> ()> source_t;

   tokenizer(source_t const & source)
      : source_(source)
   {}

   boost::optional<std::wstring> get_next()
   {
      for (;;)
      {
         boost::optional<wchar_t> t = get_next_char();
         if (!t)
            return boost::none;

         if (is_whitespace(*t))
            continue;

         switch (*t)
         {
         case L'(':
         case L')':
         case L'\\':
            return std::wstring(1, *t);
         }

         if (!is_alpha_num(*t))
            throw std::runtime_error("bad input char");

         std::wstring r(1, *t);
         for (;;)
         {
            boost::optional<wchar_t> u = get_next_char();
            if (!u)
               return r;

            if (is_alpha_num(*u))
               r += *u;
            else
            {
               last_ = *u;
               return r;
            }
         }
      }
   }

private:
   boost::optional<wchar_t> get_next_char()
   {
      if (last_)
      {
         wchar_t t = *last_;
         last_ = boost::none;
         return t;
      }

      return source_();
   }

private:
   boost::optional<wchar_t> last_;
   source_t const & source_;
};

struct preview_tokenizer
{
   preview_tokenizer(tokenizer::source_t const & source)
      : tok_(source)
   {}

   boost::optional<std::wstring> get_next()
   {
      if (last_)
      {
         std::wstring t = *last_;
         last_ = boost::none;
         return t;
      }

      return tok_.get_next();
   }

   void push_back(std::wstring const & token)
   {
      assert(!last_);
      last_ = token;
   }

private:
   boost::optional<std::wstring> last_;
   tokenizer tok_;
};